﻿'use strict';

/* 
	Top X Tracks From Date v 0.1 16/03/20
	Search n most played tracks from a given year on library. Sorting is done by play count by default.
	Duplicates by title - artist - date are removed, so it doesn't output the same tracks
	multiple times like an auto-playlist does (if you have multiple versions of the same track).
 */

// include(fb.ProfilePath + 'scripts\\SMP\\xxx-scripts\\helpers\\helpers_xxx.js');
include('remove_duplicates.js');

// Most played n Tracks from date
function do_top_tracks_from_date({
						playlistLength = 25, 
						sortBy = '$sub(99999,%play_count%)', 
						checkDuplicatesBy = ['title', 'artist', 'date'],
						year =  new Date().getFullYear() - 1, // Previous year
						forcedQuery = 'NOT (%rating% EQUAL 2 OR %rating% EQUAL 1)',
						} = {}) {
						
		// Look if target playlist already exists and clear it. Preferred to removing it, since then we could undo later...
		const playlist_name = 'Top ' + playlistLength + ' Tracks ' + year;
		const plc = plman.PlaylistCount;
		let i = 0;
		while (i < plc) {
			if (plman.GetPlaylistName(i) == playlist_name) {
				plman.ActivePlaylist = i;
				plman.UndoBackup(plman.ActivePlaylist);
				plman.ClearPlaylist(plman.ActivePlaylist);
				break;
			} else {
				i++;
			}
		}
		if (i == plc) { //if no playlist was found before
			plman.CreatePlaylist(plc, playlist_name);
			plman.ActivePlaylist = plc;
		}
		
		// Load query
		const query = '%last_played% AFTER ' + year + '-01-01 AND NOT %first_played% AFTER ' + (year + 1) + '-01-01';
		let handleList = fb.GetQueryItems(fb.GetLibraryItems(), (forcedQuery.length ? '(' + query + ') AND (' + forcedQuery + ')' : query));
		
		// Find and remove duplicates
		if (checkDuplicatesBy !== null) {
			handleList = do_remove_duplicatesV2(handleList, sortBy, checkDuplicatesBy);
		}
		
		// Filter Play counts by date
		const datesArray = fb.TitleFormat('[%played_times%]').EvalWithMetadbs(handleList);
		const datesLastFMArray = fb.TitleFormat('[%lastfm_played_times%]').EvalWithMetadbs(handleList);
		const lastPlayedArray = fb.TitleFormat('[%last_played%]').EvalWithMetadbs(handleList);
		const firstPlayedArray = fb.TitleFormat('[%first_played%]').EvalWithMetadbs(handleList);
		const playCountArray = fb.TitleFormat('[%play_count%]').EvalWithMetadbs(handleList);
		const datesArrayLength = datesArray.length;
		let dataPool = [];
		let pool = [];
		for (let i = 0; i < datesArrayLength; i++) {
			let count = 0;
			let dateArray_i = JSON.parse(datesArray[i]).concat(JSON.parse(datesLastFMArray[i])); 
			if (dateArray_i.length) { // Every entry is also an array of dates
				dateArray_i.forEach( (date) => {
					if (date.substring(0, 4) == year) {count++}
				});
			} else { // For tracks without advanced statistics
				// If first and last plays were from selected year, then all play counts too
				if (firstPlayedArray[i].substring(0, 4) == year && lastPlayedArray[i].substring(0, 4) == year) {count += playCountArray[i]}
				// Or the first play
				else if (lastPlayedArray[i].substring(0, 4) == year) {count++}
				// Or the last play
				else if (firstPlayedArray[i].substring(0, 4) == year) {count++}
				// Note any track known to have been played at selected year will be added to the pool, and sice the handle List is already
				// sorted by play Count, it will output tracks with higher total counts when they have not advanced statistics
				// being almost equivalent to 'top_tracks.js' in that case
			}
			if (count) {
				dataPool.push({idx: i, playCount: count});
			}
		}
		
		// Order by Play Count
		dataPool.sort(function (a, b) {return b.playCount - a.playCount;});
		dataPool.forEach((item) => pool.push(handleList[item.idx]))
		handleList = new FbMetadbHandleList(pool);
		
		 // Output n tracks
		handleList.RemoveRange(playlistLength, handleList.Count);
		
		// Add to playlist
		plman.InsertPlaylistItems(plman.ActivePlaylist, 0, handleList);
		
		console.log("Playlist created: " + playlist_name);
}